from bs4 import BeautifulSoup

from ...utils import str2num


def parse_html_beforeinfo(html: str):

    def parse_racers(trs):
        # preparation
        lst_tds = [tr.select('td')[::-1] for tr in trs]
        # extract
        # # 1st tds
        tds = lst_tds[0]
        boat = tds.pop().text
        tds.pop()  # skip racer figure
        name = tds.pop().select_one('a').text
        weight = tds.pop().text
        display_time = tds.pop().text
        tilt = tds.pop().text
        propeller = tds.pop().text
        parts_exchange = list(map(
            lambda li: ''.join(li.text.split()),
            tds.pop().select('ul > li')
        ))
        tds.pop()  # skip heeader-like
        temp = tds.pop()
        r = temp.text
        b = temp['class'][-1][-1] if temp.get("class", "") else ''
        # # 2nd tds
        tds = lst_tds[1]
        c = tds[0].text
        # # 3rd tds
        tds = lst_tds[2]
        st = tds[0].text
        weight_adjustment = tds[2].text
        # # 4th tds
        tds = lst_tds[3]
        rnk = tds[0].select_one('a').text

        # normalize
        name = ''.join(name.split())
        propeller = ''.join(propeller.split())
        # # to int
        boat = str2num(boat, int, boat)
        r = str2num(r, int, '')
        b = str2num(b, int, '')
        c = str2num(c, int, '')
        rnk = str2num(rnk, int, rnk)
        # # to float
        weight = str2num(weight[:-2], float, '')
        weight_adjustment = str2num(weight_adjustment, float, '')
        display_time = str2num(display_time, float, '')
        tilt = str2num(tilt, float, '')
        st = str2num(st, float, '')

        dic = {
            'name': name,
            'weight': weight,
            'weight_adjustment': weight_adjustment,
            'display_time': display_time,
            'tilt': tilt,
            'propeller': propeller,
            'parts_exchange': parts_exchange,
            'previous_race': (
                {}
                if all(map(lambda x: x == '', [r, b, c, rnk, st]))
                else
                {
                    'race': r,
                    'boat': b,
                    'course': c,
                    'ST': st,
                    "rank": rnk
                }
            )
        }

        return f'boat{boat}', dic

    def parse_display(trs):

        dic = {}

        for i, tr in enumerate(trs):
            try:
                spans = tr.select('div > span')
                dic[f'course{i+1}'] = {
                    'boat': str2num(spans[0].text, int, ''),
                    'ST': str2num(
                        spans[2].text.replace('F.', '-0.'),
                        float,
                        ''
                    )
                }
            except IndexError:
                # Case: missing some racers
                dic[f'course{i+1}'] = {
                    'boat': '',
                    'ST': '',
                }

        return dic

    def parse_weather(div):
        # extract
        time = div.select_one('div > p').text.split()[-1]
        divs = div.select('div.weather1 > div.weather1_body > div')
        direction = [
            c
            for c in divs[0].select_one('p')['class']
            if 'is-direction' in c
        ][0]
        temperature = divs[0].select('div > span')[-1].text
        weather = divs[1].select_one('div > span').text
        wind_speed = divs[2].select('div > span')[-1].text
        try:
            wind_direc = [
                c
                for c in divs[3].select_one('p')['class']
                if 'is-wind' in c
            ][0]
        except IndexError:
            wind_direc = ""
        water_temperature = divs[4].select('div > span')[-1].text
        wave_height = divs[5].select('div > span')[-1].text
        # normalize
        direction = str2num(direction.replace('is-direction', ''), int, '')
        temperature = str2num(temperature[:-1], float, '')
        wind_direc = str2num(wind_direc.replace('is-wind', ''), int, '')
        wind_speed = str2num(wind_speed[:-1], float, '')
        water_temperature = str2num(water_temperature[:-1], float, '')
        wave_height = str2num(wave_height[:-2], float, '')
        # to dict
        return {
            'direction': direction,
            'weather': weather,
            'temperature': temperature,
            'wind_direction': wind_direc,
            'wind_speed': wind_speed,
            'water_temperature': water_temperature,
            'wave_height': wave_height,
            'time': time
        }

    # make soup
    soup = BeautifulSoup(html, 'html.parser')

    # table
    grid_units = soup.select('div.grid_unit')
    table_racer = grid_units[0].select('div.table1 > table > tbody')
    table_display = grid_units[1].select('div.table1 > table > tbody > tr')
    table_weather = grid_units[1].select_one('div.weather1')

    # parse
    dic = dict([
        parse_racers(tbody.select('tr'))
        for tbody in table_racer
    ])
    dic['start_display'] = parse_display(table_display)
    dic['weather_information'] = parse_weather(table_weather)

    return dic



from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from ..certification import check_login_status, login
from ..const import IBMBRACEORJP
from ..user_information import UserInformation


def visit_ibmbraceorjp(
    user: UserInformation,
    driver: webdriver.remote.webdriver.WebDriver,
    timeout: int = 15,
) -> bool:
    """Visit and login to im.mbrace.or.jp.

    Args:
        user (UserInformation): user information
        driver (webdriver.remote.webdriver.WebDriver): webdriver
        timeout (int, optional): timeout parameter. Defaults to 15.

    Returns:
        bool: whether login is succeeded.
    """

    # check
    if not check_login_status(driver):
        login(driver, user, timeout)

    # visit
    driver.get(IBMBRACEORJP)

    # check
    return driver.title == 'トップ - BOAT RACE インターネット投票'


def get_bet_limit(
    user: UserInformation,
    driver: webdriver.remote.webdriver.WebDriver,
    timeout: int = 15
) -> int:
    """Get the amount of deposit.

    Args:
        user (UserInformation): user information.
        driver (webdriver.remote.webdriver.WebDriver): webdriver
        timeout (int, optional): timeout parameter. Defaults to 15.

    Returns:
        int: the amount of deposit.
    """

    # visit
    visit_ibmbraceorjp(user, driver, timeout)

    # get vote limit
    WebDriverWait(driver, timeout).until(EC.presence_of_element_located((By.ID, 'currentBetLimitAmount')))  # noqa
    limit = driver.find_element(By.ID, 'currentBetLimitAmount').text

    return int(limit.replace(',', ''))



from .base import BaseOperator, DriverCheckMixin
from .static import get_bet_limit


class BettingLimitCheckOperator(BaseOperator, DriverCheckMixin):

    def do(
        self,
        timeout: int = 15,
    ) -> int:
        """To check the amount of money deposited.

        Args:
            timeout (int, optional): timeout parameter. Defaults to 15.

        Raises:
            UnableActionException:
                Occurred when driver is not Chrome, Firefox or Edge.

        Returns:
            int: the amount of deposited money
        """
        self._check_driver()
        return self.__get_bet_limit(timeout=timeout,)

    def __get_bet_limit(self, timeout: int = 15) -> int:
        return get_bet_limit(
            self._user,
            self._driver,
            timeout,
        )



from bs4 import BeautifulSoup

from ...exceptions import RaceCancelledException
from ...utils import str2num
from .scrape_odds_update_time import scrape_odds_update_time


def parse_html_odds3t(html: str):

    def parse_trio(trs):
        # preparation
        lst_tds = [tr.select('td') for tr in trs]
        # extract
        dic = {}
        for tds in lst_tds:

            if 'is-fs14' in tds[0]['class']:
                # change b2
                b2 = str2num(tds[0].text, int, tds[0].text)

                for i, (b3, odds) in enumerate(zip(tds[1::3], tds[2::3])):
                    b1 = i + 1
                    b3 = str2num(b3.text, int, -1)
                    dic[f'{b1}-{b2-(b1 >= b2)}-{b3}'] = str2num(odds.text, float, odds.text)  # noqa

            else:
                # previous b2
                for i, (b3, odds) in enumerate(zip(tds[::2], tds[1::2])):
                    b1 = i + 1
                    b3 = str2num(b3.text, int, -1)
                    dic[f'{b1}-{b2-(b1 >= b2)}-{b3}'] = str2num(odds.text, float, odds.text)  # noqa

        return dic

    # make soup
    soup = BeautifulSoup(html, 'html.parser')

    # check cancel
    if '※ 該当レースは中止になりました。' in soup.text:
        raise RaceCancelledException()

    # table
    tables = soup.select('div.table1')  # probably 2 tables
    trifecta_table = tables[-1].select('table > tbody > tr')

    # parse
    dic = parse_trio(trifecta_table)
    dic["update"] = scrape_odds_update_time(soup)

    return dic



from bs4 import BeautifulSoup

from ...utils import str2num


def parse_html_racelist(html: str):

    def parse_trs(trs):
        # preparation
        lst_tds = [tr.select('td')[::-1] for tr in trs]

        # racer info
        # # extract
        tds = lst_tds[0]
        boat = tds.pop().text
        tds.pop()  # racer figure
        temp = sum(
            map(
                lambda div: ''.join(div.get_text('/').split()).split('/'),
                tds.pop().select('div')
            ),
            []  # initial
        )
        temp = [t for t in temp if t]
        racerid, clss, name, branch, birthplace, age, weight = temp

        F, L, aveST = ''.join(tds.pop().get_text(',').split()).split(',')

        global_win, global_in2nd, global_in3rd \
            = ''.join(tds.pop().get_text(',').split()).split(',')

        local_win, local_in2nd, local_in3rd \
            = ''.join(tds.pop().get_text(',').split()).split(',')

        motor_no, motor_in2nd, motor_in3rd \
            = ''.join(tds.pop().get_text(',').split()).split(',')

        boat_no, boat_in2nd, boat_in3rd \
            = ''.join(tds.pop().get_text(',').split()).split(',')

        # # normalize
        # # # to int
        boat = str2num(boat, int, boat)
        racerid = str2num(racerid, int, racerid)
        age = str2num(age[:-1], int, age)
        F = str2num(F[-1], int, F)
        L = str2num(L[-1], int, L)
        motor_no = str2num(motor_no, int, motor_no)
        boat_no = str2num(boat_no, int, boat_no)
        # # # to float
        weight = str2num(weight[:-2], float, weight[:-2])
        aveST = str2num(aveST, float, aveST)
        global_win = str2num(global_win, float, global_win)
        global_in2nd = str2num(global_in2nd, float, global_in2nd)
        global_in3rd = str2num(global_in3rd, float, global_in3rd)
        local_win = str2num(local_win, float, local_win)
        local_in2nd = str2num(local_in2nd, float, local_in2nd)
        local_in3rd = str2num(local_in3rd, float, local_in3rd)
        motor_in2nd = str2num(motor_in2nd, float, motor_in2nd)
        motor_in3rd = str2num(motor_in3rd, float, motor_in3rd)
        boat_in2nd = str2num(boat_in2nd, float, boat_in2nd)
        boat_in3rd = str2num(boat_in3rd, float, boat_in3rd)

        # drop separation
        tds.pop()

        # racer result
        races = []
        boats = []
        courses = []
        STs = []
        ranks = []

        for _ in range(14):  # TODO move 14 to another file.
            # extract
            temp = tds.pop()
            b = ''
            if temp['class'] and ('is-boatColor' in temp['class'][-1]):
                b = int(temp['class'][-1][-1])
            r = int(temp.text) if temp.text.isdigit() else ''
            c = lst_tds[1].pop().text
            c = int(c) if c.isdigit() else ''
            st = str2num(lst_tds[2].pop().text, float, '')
            rnk = lst_tds[3].pop().select_one('a').text
            rnk = int(rnk) if rnk.isdigit() else rnk
            # keep
            races.append(r)
            boats.append(b)
            courses.append(c)
            STs.append(st)
            ranks.append(rnk)

        # to dict
        dic = {
            'racerid': racerid,
            'class': clss,
            'name': name,
            'branch': branch,
            'birthplace': birthplace,
            'age': age,
            'weight': weight,
            'F': F,
            'L': L,
            'aveST': aveST,
            'global_win_pt': global_win,
            'global_in2nd': global_in2nd,
            'global_in3rd': global_in3rd,
            'local_win_pt': local_win,
            'local_in2nd': local_in2nd,
            'local_in3rd': local_in3rd,
            'motor': motor_no,
            'motor_in2nd': motor_in2nd,
            'motor_in3rd': motor_in3rd,
            'boat': boat_no,
            'boat_in2nd': boat_in2nd,
            'boat_in3rd': boat_in3rd,
            'result': [
                dict()
                if all([
                    r == '', b == '', c == '', st == '', rnk == ''
                ]) else
                {
                    'race': r,
                    'boat': b,
                    'course': c,
                    'ST': st,
                    'rank': rnk
                }
                for r, b, c, st, rnk in zip(races, boats, courses, STs, ranks)
            ]
        }

        return f'boat{boat}', dic

    # make soup
    soup = BeautifulSoup(html, 'html.parser')

    # table
    table = soup.select('div.table1.is-tableFixed__3rdadd > table > tbody')

    # parse
    dic = dict([
        parse_trs(tbody.select('tr'))
        for tbody in table
    ])
    dic['race_title'] = list(map(
        lambda s: ''.join(s.split()),
        soup.select_one('h3').text.split()  # type: ignore
    ))

    return dic



import datetime

from bs4 import BeautifulSoup

from ...const import BOATS_GEN


def parse_html_raceindex(html: str):

    def parse_tds(tds):
        # create key and extract date
        temp = tds[0]
        race = temp.select_one('a').text
        d = datetime.datetime.strptime(
            temp.select_one('a')['href'].split('hd=')[-1],
            '%Y%m%d'
        ).strftime('%Y-%m-%d')

        # extract info
        dic = {}
        dic['vote_limit'] = datetime.datetime.strptime(
            ' '.join([d, tds[1].text]),
            '%Y-%m-%d %H:%M'
        ).strftime('%Y-%m-%d %H:%M:%S')
        dic['status'] = ''.join(tds[2].text.split())
        dic['racers'] = {}
        for b in BOATS_GEN:
            divs = tds[2+b].select('div')
            dic['racers'][f'boat{b}'] = {
                'name': divs[0].select_one('a').text.replace("　", ''),
                'class': divs[-1].text
            }
        return race, dic

    # make soup
    soup = BeautifulSoup(html, 'html.parser')

    # table
    table = soup.select('div.table1 > table > tbody')

    # parse
    dic = dict([
        parse_tds(tbody.select('tr > td'))
        for tbody in table
    ])

    return dic



from datetime import date

from .const import NUM_RACES, NUM_STADIUMS


def validate_date(d: date):
    """Validate date.

    Args:
        d (date)

    Raises:
        ValueError: Occurred when d >= {date.today()}
    """
    if d > date.today():
        raise ValueError(f"Date must be <= today. But {d} is given.")


def validate_stadium(stadium: int):
    """Validate stadium ID

    Args:
        stadium (int)

    Raises:
        ValueError: Occurred when stadium is not in [1, 2, ..., {NUM_STADIUMS}]
    """
    if (stadium < 1) or (NUM_STADIUMS < stadium):
        raise ValueError(
            f'Stadium must be between 1 and {NUM_STADIUMS}. '
            f'But {stadium} is given.'
        )


def validate_race(race: int):
    """Validate race no

    Args:
        race (int)

    Raises:
        ValueError: Occurred when race is not in [1, 2, ..., {NUM_RACES}].
    """
    if (race < 1) or (NUM_RACES < race):
        raise ValueError(
            f'Race must be between 1 and {NUM_RACES}. {race} is given.'
        )



from bs4 import BeautifulSoup

from ...exceptions import RaceCancelledException
from ...utils import str2num
from .scrape_odds_update_time import scrape_odds_update_time


def parse_html_oddstf(html: str):

    def parse_win(trs):
        # preparation
        lst_tds = [tr.select('td') for tr in trs]
        # parse
        return {
            ''.join(tds[0].text.split()): str2num(tds[2].text,
                                                  float,
                                                  tds[2].text)
            for tds in lst_tds
        }

    def parse_place_show(trs):
        # preparation
        lst_tds = [tr.select('td') for tr in trs]
        # parse
        return {
            ''.join(tds[0].text.split()): list(map(
                lambda s: str2num(s, float, s),
                tds[2].text.split('-')
            ))
            for tds in lst_tds
        }

    # make soup
    soup = BeautifulSoup(html, 'html.parser')

    # check cancel
    if '※ 該当レースは中止になりました。' in soup.text:
        raise RaceCancelledException()

    # table
    grid_units = soup.select('div.grid_unit')  # probably 2 units
    win_table = grid_units[0].select('div.table1 > table > tbody > tr')
    placeshow_table = grid_units[1].select('div.table1 > table > tbody > tr')

    # parse
    dic = {
        "update": scrape_odds_update_time(soup),
        'win': parse_win(win_table),
        'place_show': parse_place_show(placeshow_table)
    }

    return dic



from logging import Logger, getLogger

from bs4 import BeautifulSoup

_logger: Logger = getLogger(__name__)

def scrape_odds_update_time(
    soup: BeautifulSoup,
    logger: Logger = _logger,
) -> str:

    # case: after time-limit
    ps = soup.select("p.tab4_time")
    if ps:
        if len(ps) >= 2:
            logger.warning(f"# of p.tab4_time should be 1 but {len(ps)}")
        p = ps[0]
        return p.text

    # case: before time-limit
    ps = soup.select("p.tab4_refreshText")
    if ps:
        if len(ps) >= 2:
            logger.warning(f"# of p.tab4_refreshText should be 1 but {len(ps)}")  # noqa
        p = ps[0]
        return p.text.split()[-1]  # Suppose that p.text == "オッズ更新時間　9:02"

    logger.warning("Failed to find the update timestamp")
    return "Failed to find the update timestamp"



from abc import ABCMeta, abstractmethod
from logging import Logger, getLogger
from typing import Any, Callable, Dict

from selenium import webdriver

_logger: Logger = getLogger(__name__)

class BaseScraper(metaclass=ABCMeta):
    """Base class for scraper.
    """

    def __init__(
        self,
        driver: webdriver.remote.webdriver.WebDriver,
        parser: Callable[[str, ], Dict[str, Any]],
        logger: Logger = _logger,
    ):
        self._driver = driver
        self._parser = parser
        self._logger = logger

    @classmethod
    @abstractmethod
    def make_url(cls, *args, **kwargs) -> str:
        raise NotImplementedError()

    def get(self, *args, **kwargs) -> Dict[str, Any]:
        url = self.make_url(*args, **kwargs)
        self._logger.info(f"URL created: {url}")
        self._driver.get(url)
        html = self._driver.page_source
        return self._parser(html)



from logging import Logger, getLogger

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from .const import BOATRACEJP_LOGIN_URL, BOATRACEJP_LOGOUT_URL, BOATRACEJP_MAIN_URL
from .exceptions import LoginFailException
from .user_information import UserInformation

# TODO error handling : failed to read
_logger: Logger = getLogger(__name__)


def login(
    driver: webdriver.remote.webdriver.WebDriver,
    user: UserInformation,
    timeout: int = 15,
    logger: Logger = _logger,
) -> bool:
    """Login to boatrace.jp

    Args:
        driver (webdriver.remote.webdriver.WebDriver): webdriver.
        user (UserInformation): User information.
        timeout (int, optional): Timeout parameter. Defaults to 15.
        logger (Logger, optional): logger. Defaults to getLogger(__name__).

    Raises:
        ValueError: Occurred when user.userid is None.
        ValueError: Occurred when user.pin is None.
        ValueError: Occurred when user.auth_pass is None.
        LoginFailException: Occurred when login failure.

    Returns:
        bool: Whether login is succeeded.

    NOTE:
        Behavior:
            Return True when login is succeeded;
            Raise LoginFailException when login is failed.
    """
    # validate
    if user.userid is None:
        raise ValueError('User ID is not set.')
    if user.pin is None:
        raise ValueError('PIN is not set.')
    if user.auth_pass is None:
        raise ValueError('Authentification password is not set.')

    # get
    driver.get(BOATRACEJP_LOGIN_URL)

    # send keys
    WebDriverWait(driver, timeout).until(
        EC.presence_of_element_located((By.NAME, 'in_KanyusyaNo'))
    )
    driver.find_element(By.CSS_SELECTOR, 'input[name="in_KanyusyaNo"]')\
          .send_keys(user.userid)
    logger.debug('put userid')

    WebDriverWait(driver, timeout).until(
        EC.presence_of_element_located((By.NAME, 'in_AnsyoNo'))
    )
    driver.find_element(By.CSS_SELECTOR, 'input[name="in_AnsyoNo"]')\
          .send_keys(user.pin)
    logger.debug('put user pin')

    WebDriverWait(driver, timeout).until(
        EC.presence_of_element_located((By.NAME, 'in_PassWord'))
    )
    driver.find_element(By.CSS_SELECTOR, 'input[name="in_PassWord"]')\
          .send_keys(user.auth_pass)
    logger.debug('put authentification password')

    # press button
    WebDriverWait(driver, timeout).until(
        EC.presence_of_element_located((By.CLASS_NAME, 'is-type3_2'))
    )
    driver.find_element(By.CSS_SELECTOR, 'button[class="btn is-type3_2"]')\
          .click()
    logger.debug('Pressed login button')

    is_successed = check_login_status(driver)

    if not is_successed:
        raise LoginFailException()

    return is_successed


def logout(
    driver: webdriver.remote.webdriver.WebDriver,
    logger: Logger = _logger,
) -> bool:
    """Logout from boatrace.jp

    Args:
        driver (webdriver.remote.webdriver.WebDriver): webdriver
        logger (Logger, optional): logger. Defaults to getLogger(__name__).

    Returns:
        bool: Whether logout is succeeded.
    """
    # logout
    driver.get(BOATRACEJP_LOGOUT_URL)
    return not check_login_status(driver)


def check_login_status(
    driver: webdriver.remote.webdriver.WebDriver,
    logger: Logger = _logger,
) -> bool:
    """[summary]

    Args:
        driver (webdriver.remote.webdriver.WebDriver): webdriver
        logger (Logger, optional): logger. Defaults to getLogger(__name__).

    Returns:
        bool: Whether you are loggined in boatrace.jp
    """
    # get
    driver.get(BOATRACEJP_MAIN_URL)

    if driver.find_elements(By.CLASS_NAME, 'is-logout1'):
        return True
    else:
        return False



import json
import os
from getpass import getpass
from logging import Logger, getLogger
from typing import Optional

_logger: Logger = getLogger(__name__)


class UserInformation:
    """ Use-information container class

    - userid
    - pin
    - auth_pass
    - vote_pass

    """

    def __init__(
        self,
        userid: Optional[str] = None,
        pin: Optional[str] = None,
        auth_pass: Optional[str] = None,
        vote_pass: Optional[str] = None,  # TODO rename vote_pass -> bet_pass
        json_file: Optional[str] = None,
        logger: Logger = _logger,
    ):
        # preparation
        self.userid: Optional[str] = None
        self.pin: Optional[str] = None
        self.auth_pass: Optional[str] = None
        self.vote_pass: Optional[str] = None

        # load json
        if json_file is not None and os.path.exists(json_file):
            with open(json_file, 'r', encoding='utf-8-sig') as f:
                dic = json.load(f)
            self.userid = dic.get('userid', None)
            self.pin = dic.get('pin', None)
            self.auth_pass = dic.get('auth_pass', None)
            self.vote_pass = dic.get('vote_pass', None)

        if userid is not None:
            self.userid = userid
        if pin is not None:
            self.pin = pin
        if auth_pass is not None:
            self.auth_pass = auth_pass
        if vote_pass is not None:
            self.vote_pass = vote_pass

        if self.userid is None:
            self.userid = input('Input your id: ')
        if self.pin is None:
            self.pin = getpass(u'Input pin (暗証番号): ')
        if self.auth_pass is None:
            self.auth_pass = getpass(
                u'Input authentification password (認証用パスワード): '
            )
        if self.vote_pass is None:
            self.vote_pass = getpass(
                u'Input vote password (投票用パスワード): '
            )



import re
import time
from datetime import date, datetime, timedelta
from typing import Any, Dict
from unittest import mock

import pytest

from pyjpboatrace import PyJPBoatrace
from pyjpboatrace.const import STADIUMS_MAP
from pyjpboatrace.exceptions import NoDataException, RaceCancelledException

from ._driver_fixutures import chrome_driver  # noqa
from ._utils import (
    get_expected_json,
    get_mock_html,
    get_user_info,
    is_boatrace_time,
)

# TODO add test for get function of racer's basic info
# TODO add test for get function of racer's last 3sections info
# TODO add test for get function of racer's season info
# TODO add test for get function of racer's course-wise info


@pytest.fixture(scope='module')
def boatrace_tools(chrome_driver) -> PyJPBoatrace:  # type: ignore  # noqa
    user = get_user_info()
    pyjpboatrace = PyJPBoatrace(
        driver=chrome_driver,
        user_information=user,
        close_driver_when_closing_pyjpboatrace=False,
        # because closing driver in pytest.fixture
    )
    yield pyjpboatrace
    pyjpboatrace.close()


@pytest.mark.integrate
@pytest.mark.parametrize(
    "d",
    [
        date(2020, 9, 8),
        date(2022, 10, 30),
    ]
)
def test_get_stadiums(d: date, boatrace_tools: PyJPBoatrace):
    # preparation
    dstr = d.strftime('%Y%m%d')
    # expectation
    expected = get_expected_json(f"expected_index.hd={dstr}.json")
    # actual
    actual = boatrace_tools.get_stadiums(d)
    # assertion
    assert actual == expected


@mock.patch('selenium.webdriver.Chrome')
def test_get_stadiums_today(mock_chrome):
    # TODAY (=2020/11/30) CASE #
    # preparation
    d = date(2020, 11, 30)
    # set mock
    mock_chrome.page_source = get_mock_html("today_index.html")

    # expectation
    expected = get_expected_json('expected_today_index.json')
    expected.update(date=d.strftime("%Y-%m-%d"))

    # actual
    actual = PyJPBoatrace(driver=mock_chrome).get_stadiums(d)

    # assert
    assert actual == expected

    # close
    mock_chrome.close()


@pytest.mark.integrate
def test_get_12races(boatrace_tools: PyJPBoatrace):

    # preparation
    d = date(2020, 10, 8)
    dstr = d.strftime('%Y%m%d')
    stadium = 1

    # expectation
    expected = get_expected_json(
        f"expected_raceindex.jcd={stadium:02d}&hd={dstr}.json",
    )

    # actual
    actual = boatrace_tools.get_12races(d, stadium)

    # assertion
    assert actual == expected


@mock.patch('selenium.webdriver.Chrome')
def test_get_12races_today(mock_chrome):
    # TODAY (=2020/12/01) CASE #
    # preparation
    d = date(2020, 12, 1)
    stadium = 1
    # set mock
    mock_chrome.page_source = get_mock_html("today_raceindex.html")

    # expectation
    expected = get_expected_json('expected_today_raceindex.json')
    expected.update(date=d.strftime("%Y-%m-%d"), stadium=stadium)

    # actual
    actual = PyJPBoatrace(driver=mock_chrome).get_12races(d, stadium)

    # assert
    assert actual == expected

    # close
    mock_chrome.close()


@pytest.mark.integrate
def test_get_race_info(boatrace_tools: PyJPBoatrace):

    # USUAL CASE #
    # preparation
    d = date(2020, 10, 24)
    dstr = d.strftime('%Y%m%d')
    stadium = 14
    race = 1
    # expectation
    expected = get_expected_json(
        f'expected_racelist.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual
    actual = boatrace_tools.get_race_info(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_race_info_missing_racer(boatrace_tools: PyJPBoatrace):
    # MISSING RACERS CASE #
    # preparation
    d = date(2020, 11, 29)
    dstr = d.strftime('%Y%m%d')
    stadium = 10
    race = 2
    # expectation
    expected = get_expected_json(
        f'expected_racelist.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual
    actual = boatrace_tools.get_race_info(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_odds_win_placeshow(boatrace_tools: PyJPBoatrace):
    # USUAL CASE #
    # preparation
    d = date(2020, 10, 24)
    dstr = d.strftime('%Y%m%d')
    stadium = 14
    race = 1
    # load true data
    expected = get_expected_json(
        f'expected_oddstf.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual data
    actual = boatrace_tools.get_odds_win_placeshow(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_odds_win_placeshow_missing_racer(boatrace_tools: PyJPBoatrace):
    # MISSING RACERS CASE #
    # preparation
    d = date(2020, 11, 29)
    dstr = d.strftime('%Y%m%d')
    stadium = 10
    race = 2
    # expectation
    expected = get_expected_json(
        f'expected_oddstf.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual
    actual = boatrace_tools.get_odds_win_placeshow(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_odds_win_placeshow_cancelled_race(boatrace_tools: PyJPBoatrace):
    # CANCELLED RACERS CASE #
    # preparation
    d = date(2019, 1, 26)
    stadium = 8
    race = 8
    # actual
    with pytest.raises(RaceCancelledException):
        boatrace_tools.get_odds_win_placeshow(d, stadium, race)


@pytest.mark.integrate
def test_get_odds_quinellaplace(boatrace_tools: PyJPBoatrace):
    # USUAL CASE #
    # preparation
    d = date(2020, 10, 24)
    dstr = d.strftime('%Y%m%d')
    stadium = 14
    race = 1
    # load true data
    expected = get_expected_json(
        f'expected_oddsk.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual data
    actual = boatrace_tools.get_odds_quinellaplace(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_odds_quinellaplace_missing_racer(boatrace_tools: PyJPBoatrace):
    # MISSING RACERS CASE #
    # preparation
    d = date(2020, 11, 29)
    dstr = d.strftime('%Y%m%d')
    stadium = 10
    race = 2
    # expectation
    expected = get_expected_json(
        f'expected_oddsk.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual data
    actual = boatrace_tools.get_odds_quinellaplace(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_odds_quinellaplace_cancelled_race(boatrace_tools: PyJPBoatrace):
    # CANCELLED RACE CASE #
    # preparation
    d = date(2019, 1, 26)
    stadium = 8
    race = 8
    # assert
    with pytest.raises(RaceCancelledException):
        boatrace_tools.get_odds_quinellaplace(d, stadium, race)


@pytest.mark.integrate
def test_get_odds_exacta_quinella(boatrace_tools: PyJPBoatrace):
    # USUAL CASE #
    # preparation
    d = date(2020, 10, 24)
    dstr = d.strftime('%Y%m%d')
    stadium = 14
    race = 1
    # load true data
    expected = get_expected_json(
        f'expected_odds2tf.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual data
    actual = boatrace_tools.get_odds_exacta_quinella(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_odds_exacta_quinella_missing_racer(boatrace_tools: PyJPBoatrace):
    # MISSING RACERS CASE #
    # preparation
    d = date(2020, 11, 29)
    dstr = d.strftime('%Y%m%d')
    stadium = 10
    race = 2
    # expectation
    expected = get_expected_json(
        f'expected_odds2tf.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual data
    actual = boatrace_tools.get_odds_exacta_quinella(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_odds_exacta_quinella_cancelled_race(boatrace_tools: PyJPBoatrace):
    # CANCELLED RACE CASE #
    # preparation
    d = date(2019, 1, 26)
    stadium = 8
    race = 8
    # assert
    with pytest.raises(RaceCancelledException):
        boatrace_tools.get_odds_exacta_quinella(d, stadium, race)


@pytest.mark.integrate
def test_get_odds_trifecta(boatrace_tools: PyJPBoatrace):
    # USUAL CASE #
    # preparation
    d = date(2020, 10, 24)
    dstr = d.strftime('%Y%m%d')
    stadium = 14
    race = 1
    # load true data
    expected = get_expected_json(
        f'expected_odds3t.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )

    # actual data
    actual = boatrace_tools.get_odds_trifecta(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_odds_trifecta_missing_racer(boatrace_tools: PyJPBoatrace):
    # MISSING RACERS CASE #
    # preparation
    d = date(2020, 11, 29)
    dstr = d.strftime('%Y%m%d')
    stadium = 10
    race = 2
    # load true data
    expected = get_expected_json(
        f'expected_odds3t.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual data
    actual = boatrace_tools.get_odds_trifecta(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_odds_trifecta_cancelled_race(boatrace_tools: PyJPBoatrace):
    # CANCELLED RACERS CASE #
    # preparation
    d = date(2019, 1, 26)
    stadium = 8
    race = 8
    # assert
    with pytest.raises(RaceCancelledException):
        boatrace_tools.get_odds_trifecta(d, stadium, race)


@pytest.mark.integrate
def test_get_odds_trio(boatrace_tools: PyJPBoatrace):
    # USUAL CASE #
    # preparation
    d = date(2020, 10, 24)
    dstr = d.strftime('%Y%m%d')
    stadium = 14
    race = 1
    # load true data
    expected = get_expected_json(
        f'expected_odds3f.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual data
    actual = boatrace_tools.get_odds_trio(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
@pytest.mark.skipif(
    not is_boatrace_time(),
    reason='it is not time for boatrace'
)
@pytest.mark.parametrize(
    "method_name",
    [
        "get_odds_trifecta",
        "get_odds_trio",
        "get_odds_exacta_quinella",
        "get_odds_quinellaplace",
        "get_odds_win_placeshow",
    ]
)
def test_get_real_time_odds(method_name: str, boatrace_tools: PyJPBoatrace):

    # preparation
    today = date.today()
    stadiums = boatrace_tools.get_stadiums(today)

    # search active race
    for stadium_str, dic in stadiums.items():
        if stadium_str not in (s for _, s in STADIUMS_MAP):
            continue
        if "R以降発売中" in dic["status"]:
            race = dic["next_race"]
            stadium = {k: v for v, k in STADIUMS_MAP}.get(stadium_str)
            break
    else:
        raise Exception("Failed to find any active race.")

    # actual data
    actual: Dict[str, Any]
    actual = boatrace_tools.__getattribute__(method_name)(today, stadium, race)
    # assertion
    assert actual.get("date") == today.strftime("%Y-%m-%d")
    assert actual.get("stadium") == stadium
    assert actual.get("race") == race
    assert re.fullmatch(r"(\d|\d\d):\d\d", actual.get("update"))  # type: ignore  # noqa


@pytest.mark.integrate
def test_get_odds_trio_missing_racer(boatrace_tools: PyJPBoatrace):
    # MISSING RACERS CASE #
    # preparation
    d = date(2020, 11, 29)
    dstr = d.strftime('%Y%m%d')
    stadium = 10
    race = 2
    # expectation
    expected = get_expected_json(
        f'expected_odds3f.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual
    actual = boatrace_tools.get_odds_trio(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_odds_trio_cancelled_race(boatrace_tools: PyJPBoatrace):
    # CANCELLED RACERS CASE #
    # preparation
    d = date(2019, 1, 26)
    stadium = 8
    race = 8
    # assert
    with pytest.raises(RaceCancelledException):
        boatrace_tools.get_odds_trio(d, stadium, race)


@pytest.mark.integrate
def test_get_just_before_info(boatrace_tools: PyJPBoatrace):
    # USUAL CASE #
    # preparation
    d = date(2020, 8, 25)
    dstr = d.strftime('%Y%m%d')
    stadium = 14
    race = 7
    # load true data
    expected = get_expected_json(
        f'expected_beforeinfo.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual data
    actual = boatrace_tools.get_just_before_info(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_just_before_info_missing_racer(boatrace_tools: PyJPBoatrace):
    # MISSING RACERS CASE #
    # preparation
    d = date(2020, 11, 29)
    dstr = d.strftime('%Y%m%d')
    stadium = 10
    race = 2
    # expectation
    expected = get_expected_json(
        f'expected_beforeinfo.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual
    actual = boatrace_tools.get_just_before_info(d, stadium, race)
    # assertion
    assert actual == expected


@mock.patch('selenium.webdriver.Chrome')
def test_get_just_before_info_not_yet(mock_chrome):
    # NOT YET DISPLAYED CASE#
    # preparation: anything OK
    d = date(2020, 11, 29)
    stadium = 10
    race = 2
    # set mock
    mock_chrome.page_source = get_mock_html("not_yet_beforeinfo.html")

    # expectation
    expected = get_expected_json("expected_not_yet_beforeinfo.json")
    expected.update(date=d.strftime("%Y-%m-%d"), stadium=stadium, race=race)
    # actual
    pyjpboatrace = PyJPBoatrace(driver=mock_chrome)
    actual = pyjpboatrace.get_just_before_info(d, stadium, race)
    # assertion
    assert actual == expected
    # close
    mock_chrome.close()


@pytest.mark.integrate
def test_get_just_before_info_cancelled_race(boatrace_tools: PyJPBoatrace):
    # CANCELLED RACERS CASE #
    # preparation
    d = date(2019, 1, 26)
    dstr = d.strftime('%Y%m%d')
    stadium = 8
    race = 8
    # expectation
    expected = get_expected_json(
        f'expected_beforeinfo.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual
    actual = boatrace_tools.get_just_before_info(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_race_result(boatrace_tools: PyJPBoatrace):
    # USUAL CASE #
    # preparation
    d = date(2020, 10, 24)
    dstr = d.strftime('%Y%m%d')
    stadium = 14
    race = 1
    # load true data
    expected = get_expected_json(
        f'expected_raceresult.rno={race}&jcd={stadium:02d}&hd={dstr}.json',
    )
    # actual data
    actual = boatrace_tools.get_race_result(d, stadium, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
@pytest.mark.parametrize(
    "d,std,race",
    [
        (date(2020, 11, 29), 10, 2),
        (date(2018, 1, 1), 21, 3),
        # NOTE: https://boatrace.jp/owpc/pc/race/raceresult?rno=10&jcd=01&hd=20130922 has not responsed data since 2023/10/08.  # noqa
        # (date(2013, 9, 22), 1, 10),
    ]
)
def test_get_race_result_missing_racer(d, std, race, boatrace_tools):
    # MISSING RACERS CASE #
    # preparation
    dstr = d.strftime('%Y%m%d')
    # expectation
    expected = get_expected_json(
        f'expected_raceresult.rno={race}&jcd={std:02d}&hd={dstr}.json',
    )
    # actual
    actual = boatrace_tools.get_race_result(d, std, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
@pytest.mark.parametrize(
    "d,std,race",
    [
        (date(2019, 4, 8), 11, 5),
        (date(2021, 6, 14), 12, 1),
        (date(2021, 1, 3), 14, 10),
    ]
)
def test_get_race_result_for_tie_case(d, std, race, boatrace_tools):
    # MISSING RACERS CASE #
    # preparation
    dstr = d.strftime('%Y%m%d')
    # expectation
    expected = get_expected_json(
        f'expected_raceresult.rno={race}&jcd={std:02d}&hd={dstr}.json',
    )
    # actual
    actual = boatrace_tools.get_race_result(d, std, race)
    # assertion
    assert actual == expected


@pytest.mark.integrate
def test_get_race_result_cancelled_race(boatrace_tools: PyJPBoatrace):
    # CANCELLED RACERS CASE #
    # preparation
    d = date(2019, 1, 26)
    stadium = 8
    race = 8
    # assert
    with pytest.raises(RaceCancelledException):
        boatrace_tools.get_race_result(d, stadium, race)


@mock.patch('selenium.webdriver.Chrome')
def test_get_race_result_not_yet(mock_chrome):
    # NOT YET DISPLAYED CASE#
    # preparation: anything OK
    d = date(2020, 11, 29)
    stadium = 10
    race = 2
    # set mock
    mock_chrome.page_source = get_mock_html("not_yet_raceresult.html")

    # assert
    pyjpboatrace = PyJPBoatrace(driver=mock_chrome)
    with pytest.raises(NoDataException):
        pyjpboatrace.get_race_result(d, stadium, race)

    # close
    mock_chrome.close()


@pytest.mark.integrate
def test_get_race_result_invalid_arguments(boatrace_tools: PyJPBoatrace):

    # TODO invalid args test for get_stadiums
    # TODO invalid args test for get_12races
    # TODO invalid args test for get_just_before_info
    # TODO invalid args test for get_race_info
    # TODO invalid args test for get_odds_win_placeshow
    # TODO invalid args test for get_odds_quinellaplace
    # TODO invalid args test for get_odds_exacta_quinella
    # TODO invalid args test for get_odds_trifecta
    # TODO invalid args test for get_odds_trio

    # invalid date
    d = (datetime.today()+timedelta(days=1)).date()
    stadium = 10
    race = 2
    # msg = f'Date d must be before today. {d} is given.'
    with pytest.raises(ValueError):
        boatrace_tools.get_race_result(d, stadium, race)

    # invalid stadium
    d = datetime.today().date()
    stadium = 0
    race = 2
    # msg = f'Stadium must be between 1 and 24. {stadium} is given.'
    with pytest.raises(ValueError):
        boatrace_tools.get_race_result(d, stadium, race)

    # invalid race
    d = datetime.today().date()
    stadium = 10
    race = 13
    # msg = f'Race must be between 1 and 12. {race} is given.'
    with pytest.raises(ValueError):
        boatrace_tools.get_race_result(d, stadium, race)


@pytest.mark.skipif(
    not is_boatrace_time(),
    reason='it is not time for boatrace'
)
@pytest.mark.skipif(
    get_user_info() is None,
    reason='UserInformation not found'
)
@pytest.mark.integrate
@pytest.mark.spending_money
def test_deposit_withdraw(boatrace_tools: PyJPBoatrace):
    # pre-status
    current = boatrace_tools.get_bet_limit()
    # deposit
    num = 1000
    boatrace_tools.deposit(num//1000)
    time.sleep(10)
    after = boatrace_tools.get_bet_limit()
    assert after == current+num
    # withdraw
    boatrace_tools.withdraw()
    time.sleep(10)
    current = boatrace_tools.get_bet_limit()
    assert current == 0


@pytest.mark.skipif(
    not is_boatrace_time(),
    reason='it is not time for boatrace'
)
@pytest.mark.skipif(
    get_user_info() is None,
    reason='UserInformation not found'
)
@pytest.mark.integrate
@pytest.mark.spending_money
def test_bet(boatrace_tools: PyJPBoatrace):
    # preparation
    current = boatrace_tools.get_bet_limit()
    if current <= 700:
        # deposit
        num = 1000
        boatrace_tools.deposit(num//1000)
        time.sleep(10)

    # scrape stadiums/races
    stadiums_dic = boatrace_tools.get_stadiums(date.today())

    _flag = False
    for key in stadiums_dic:
        stadium = {s: i for i, s in STADIUMS_MAP}[key]
        races_dic = boatrace_tools.get_12races(date.today(), stadium)
        for r, dic in races_dic.items():
            if dic.get("status") == "投票":
                race = int(r[:-1])
                _flag = True
                break
        if _flag:
            break
        time.sleep(2)

    if not _flag:
        raise Exception("Failed to find any active races.")

    # bet list
    betdict = {
        'trifecta': {'1-2-4': 100},
        'trio': {'2=3=4': 100},
        'exacta': {'2-1': 100},
        'quinella': {'3=4': 100},
        'quinellaplace': {'2=6': 100},
        'win': {'5': 100},
        'placeshow': {'6': 100},
    }

    assert boatrace_tools.bet(
        stadium=stadium,
        race=race,
        trifecta_betting_dict=betdict['trifecta'],
        trio_betting_dict=betdict['trio'],
        exacta_betting_dict=betdict['exacta'],
        quinella_betting_dict=betdict['quinella'],
        quinellaplace_betting_dict=betdict['quinellaplace'],
        win_betting_dict=betdict['win'],
        placeshow_betting_dict=betdict['placeshow']
    )



class NoDataException(Exception):
    pass


class RaceCancelledException(Exception):
    def __init__(self, msg=None):
        super().__init__(msg)


class UnableActionException(Exception):
    def __init__(self, msg=None):
        super().__init__(msg)


class LoginFailException(Exception):
    pass


class InsufficientDepositException(Exception):
    pass


class ZeroDepositException(InsufficientDepositException):
    pass


class VoteNotInTimeException(Exception):
    pass


class InactiveStadium(Exception):
    pass


class InactiveRace(Exception):
    pass


class UserInformationNotGivenException(Exception):
    pass



import datetime
from logging import Logger, getLogger
from typing import Any, Dict

from selenium import webdriver

from ..const import BOATRACEJP_BASE_URL
from ..validator import validate_date
from ._parser import parse_html_index
from .base import BaseScraper

_logger: Logger = getLogger(__name__)

class StadiumsScraper(BaseScraper):
    """To get stadiums where races are/were held.
    """

    __url_format = f"{BOATRACEJP_BASE_URL}/index?hd={{date}}"

    def __init__(
        self,
        driver: webdriver.remote.webdriver.WebDriver,
        logger: Logger = _logger,
    ):
        super().__init__(driver, parse_html_index, logger)

    @classmethod
    def make_url(cls, d: datetime.date) -> str:
        """Make target URL.

        Args:
            d (datetime.date): race date

        Returns:
            str: URL
        """
        return cls.__url_format.format(date=d.strftime("%Y%m%d"))

    def get(self, d: datetime.date) -> Dict[str, Any]:
        """Get stadiums where races are held on the given day.

        Args:
            d (datetime.date): race date

        Raises:
            ValueError: Occurred when invalid date given.

        Returns:
            Dict[str, Any]: scraped data
        """
        validate_date(d)
        dic = super().get(d)
        dic.update(
            date=d.strftime("%Y-%m-%d"),
        )
        return dic



from . import (
    certification,
    const,
    drivers,
    exceptions,
    operator,
    pyjpboatrace,
    scraper,
    user_information,
    utils,
    validator,
)
from .pyjpboatrace import PyJPBoatrace

__version__ = 'v0.4.1'

__all__ = [
    PyJPBoatrace.__name__,
    certification.__name__,
    const.__name__,
    drivers.__name__,
    exceptions.__name__,
    operator.__name__,
    pyjpboatrace.__name__,
    scraper.__name__,
    user_information.__name__,
    utils.__name__,
    validator.__name__,
]



from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from ..exceptions import ZeroDepositException
from .base import BaseOperator, DriverCheckMixin
from .static import get_bet_limit, visit_ibmbraceorjp


class WithdrawOperator(BaseOperator, DriverCheckMixin):

    def do(
        self,
        timeout: int = 15,
        raise_validation_error: bool = False,
    ) -> bool:
        """To withdraw deposit.

        Args:
            timeout (int, optional): timeout parameter. Defaults to 15.
            raise_validation_error (bool, optional): whether raise validation error. Defaults to False.

        Raises:
            ValueError:
                Occurred when raise_validation_error is True and the vote password is not set.
            ZeroDepositException:
                Occurred when no deposit.
            UnableActionException:
                Occurred when driver is not Chrome, Firefox or Edge.
        """  # noqa
        self._check_driver()
        return self.__withdraw(
            timeout=timeout,
            raise_validation_error=raise_validation_error,
        )

    def __withdraw(
        self,
        timeout: int = 15,
        raise_validation_error: bool = False,
    ):
        # validation
        if self._user.vote_pass is None:
            if raise_validation_error:
                raise ValueError('Vote password is not set.')
            self._logger.error('Vote password is not set.')
            return None

        # get current bet limit
        current_limit = get_bet_limit(
            self._user,
            self._driver,
            timeout,
        )
        if current_limit == 0:
            # TODO add test
            raise ZeroDepositException('Current deposit is zero.')

        # visit
        visit_ibmbraceorjp(self._user, self._driver, timeout)

        # click deposit/withdraw
        WebDriverWait(self._driver, timeout).until(EC.presence_of_element_located((By.ID, 'gnavi01')))  # noqa
        self._driver.find_element(By.ID, 'gnavi01').click()

        WebDriverWait(self._driver, timeout).until(EC.presence_of_element_located((By.ID, 'account')))  # noqa
        self._driver.find_element(By.ID, 'account').click()

        # input
        WebDriverWait(self._driver, timeout).until(EC.presence_of_element_located((By.ID, 'accountBetPassword')))  # noqa
        self._driver.find_element(By.ID, 'accountBetPassword')\
            .send_keys(self._user.vote_pass)

        # press button
        WebDriverWait(self._driver, timeout).until(EC.presence_of_element_located((By.ID, 'executeAccount')))  # noqa
        self._driver.find_element(By.ID, 'executeAccount')\
            .click()

        WebDriverWait(self._driver, timeout).until(EC.presence_of_element_located((By.ID, 'ok')))  # noqa
        self._driver.find_element(By.ID, 'ok')\
            .click()



from datetime import time

# TODO make them constant

# URL
# TODO format
BOATRACEJP_MAIN_URL = 'https://www.boatrace.jp/'
BOATRACEJP_LOGIN_URL = f'{BOATRACEJP_MAIN_URL}owpc/pc/login?authAfterUrl=/'
BOATRACEJP_LOGOUT_URL = f'{BOATRACEJP_MAIN_URL}owpc/logout'
BOATRACEJP_BASE_URL = f"{BOATRACEJP_MAIN_URL}owpc/pc/race"


IBMBRACEORJP = ''.join([
    f'{BOATRACEJP_MAIN_URL}',
    'owpc/VoteBridgeNew.jsp?',
    'param=H0JS00000stContens'
    '&kbn=1'
    '&voteActionUrl=/owpc/pc/site/index.html'
])

# TIME
BOATRACE_START = time(hour=8, minute=30)  # 8:30
BOATRACE_END = time(hour=20, minute=45)  # 20:45

# RACE
# # races
NUM_RACES = 12  # # of races in a stadium on a day
# # boats
NUM_BOATS = 6
BOATS_GEN = range(1, NUM_BOATS+1)
# # stadiums
STADIUMS_MAP = tuple((i+1, stadium) for i, stadium in enumerate([
    "桐生",
    "戸田",
    "江戸川",
    "平和島",
    "多摩川",
    "浜名湖",
    "蒲郡",
    "常滑",
    "津",
    "三国",
    "びわこ",
    "住之江",
    "尼崎",
    "鳴門",
    "丸亀",
    "児島",
    "宮島",
    "徳山",
    "下関",
    "若松",
    "芦屋",
    "福岡",
    "唐津",
    "大村",
]))
NUM_STADIUMS = len(STADIUMS_MAP)



from .base import BaseOperator, DriverCheckMixin
from .better import BettingOperator
from .betting_limit_checker import BettingLimitCheckOperator
from .depositor import DepositOperator
from .withdrawer import WithdrawOperator

__all__ = [
    BaseOperator.__name__,
    DriverCheckMixin.__name__,
    BettingOperator.__name__,
    BettingLimitCheckOperator.__name__,
    DepositOperator.__name__,
    WithdrawOperator.__name__,
]
